use std::fmt::{Debug, Display};

use anyhow::{bail, Result};
use reqwest::Response;

pub struct WithIndex<T> {
    value: T,
    index: usize,
}

impl<T> WithIndex<T> {
    pub fn index_from(resp: &Response) -> Result<WithIndexBuilder<T>> {
        let index = match resp.headers().get("X-Consul-Index") {
            Some(v) => v.to_str()?.parse::<usize>()?,
            None => bail!("X-Consul-Index header not found"),
        };
        Ok(WithIndexBuilder {
            index,
            _phantom: Default::default(),
        })
    }

    pub fn into_inner(self) -> T {
        self.value
    }

    pub fn index(&self) -> usize {
        self.index
    }
}

impl<T> std::convert::AsRef<T> for WithIndex<T> {
    fn as_ref(&self) -> &T {
        &self.value
    }
}

impl<T> std::borrow::Borrow<T> for WithIndex<T> {
    fn borrow(&self) -> &T {
        &self.value
    }
}

impl<T> std::ops::Deref for WithIndex<T> {
    type Target = T;
    fn deref(&self) -> &T {
        &self.value
    }
}

impl<T: Debug> Debug for WithIndex<T> {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        <T as Debug>::fmt(self, f)
    }
}

impl<T: Display> Display for WithIndex<T> {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        <T as Display>::fmt(self, f)
    }
}

pub struct WithIndexBuilder<T> {
    _phantom: std::marker::PhantomData<T>,
    index: usize,
}

impl<T> WithIndexBuilder<T> {
    pub fn value(self, value: T) -> WithIndex<T> {
        WithIndex {
            value,
            index: self.index,
        }
    }
}
